# -*- coding: utf-8 -*-

"""Encapsualtes a set of python specific name conversion operations.

"""
import tornado.template as template

from pyesdoc.mp.ontologies.generators import generator_utils as gu
from pyesdoc.mp.ontologies.generators.generator_utils import *



# Language name constant.
LANGUAGE = 'python'

# Language prefix constant.
LANGUAGE_PREFIX = 'py'

# Language file extension constant.
FILE_EXTENSION = '.py'

# Python package initialisation file name.
_PACKAGE_INIT_FILE = '__init__'

# Python clas property field prefix.
_PROPERTY_FIELD_PREFIX = 'self.'

# Set of simple type mappings.
_SIMPLE_TYPE_MAPPINGS = {
    'bool' : 'bool',
    'date' : 'datetime.date',
    'datetime' : 'datetime.datetime',
    'float' : 'float',
    'int' : 'int',
    'str' : 'unicode',
    'unicode' : 'unicode',
    'uri' : 'unicode',
    'uuid' : 'uuid.UUID',
}

# Standard null value.
_NULL_VALUE = 'None'

# Iterable null value.
_NULL_VALUE_COLLECTION = '[]'


def _strip(name):
    """Returns stripped name.

    Keyword Arguments:
    name - name being converted.

    """
    if isinstance(name, str) == False:
        name = name.name
    return name


def _strip_type_name(name):
    """Returns stripped type name.

    :param str name: Name being converted.

    """
    name = _strip(name)
    if name.find('.') != -1:
        name = name.split('.')[len(name.split('.')) - 1]
    return name


def get_ontology_name(name):
    """Converts name to a python ontology name.

    Keyword Arguments:
    name - name being converted.

    """
    if isinstance(name, str) == False:
        name = name.name

    return name.lower()


def get_ontology_version(name):
    """Converts version identifier to a python ontology version.

    Keyword Arguments:
    name - name of version identifier being converted.

    """
    if isinstance(name, str) == False:
        name = name.version

    return name.replace(".", "_")


def _get_ontology_directory(o, root=None, sub=None, suffix_root=False):
    """Returns ontology directory into which code is generated code.

    Keyword Arguments:
    ontology - ontology being processed.
    root - root directory with which ontology is associated.
    sub - sub directory to append as a suffix.
    suffix_root_dir - flag indicating whether to append a standard suffix to root directory.

    """
    dir = ''
    if root is not None:
        dir += root + '/'
    if suffix_root == True:
        dir += 'cim_codegen/python/'
        dir += get_ontology_name(o)
    dir += get_ontology_name(o)
    dir += '/'
    dir += 'v' + get_ontology_version(o)
    if sub is not None:
        dir += '/' + sub
    return dir


def get_ontology_directory(ctx, sub=None, include_version=True):
    """Returns ontology directory into which code is generated code.

    :param GeneratorContext ctx: Generation context information.
    :param str sub: Subpackage name.
    :param bool include_version: Subpackage name.

    """
    dir = ''
    if ctx.output_dir is not None:
        dir += ctx.output_dir + '/'
    dir += get_ontology_name(ctx.ontology)
    if include_version:
        dir += '/'
        dir += 'v' + get_ontology_version(ctx.ontology)
    if sub is not None:
        dir += '/'
        dir += sub

    return dir


def get_package_name(name):
    """Converts name to a python package name.

    :param str name: Package name being converted.

    """
    return _strip_package_name(name)


def get_full_type_name(c):
    """Converts name to a python type name.

    :param str name: Class name being converted.

    """
    return get_package_name(c.package) + "." + get_type_name(c)


def get_type_name(name):
    """Converts name to a python type name.

    :param str name: Class name being converted.

    """
    return convert_to_camel_case(_strip_type_name(name))


def get_type_import_name(name):
    """Converts name to a python type import name.

    :param str name: Class name being converted.

    """
    return _strip_type_name(name)


def get_type_func_name(name):
    """Converts name to one suitable for use in a python function definition.

    :param str name: Class name being converted.

    """
    return _strip_type_name(name)


def get_type_doc_string_name(name):
    """Converts name to one suitable for use in python documentation.

    :param str name: Type name being converted.

    """
    return _strip_type_name(name).replace('_', ' ')


def _get_class_base_name(c):
    """Converts name to a python base class name.

    """
    if c.base is None:
        return 'object'
    elif c.base.package == c.package:
        return get_type_name(c.base)
    else:
        return get_full_type_name(c.base)


def get_property_ctor(p):
    """Converts class property to a python property constructor declaration.

    """
    return 'self.{0} = {1}'.format(get_property_name(p),
                                   get_property_default_value(p))


def get_property_reference_ctor(p):
    """Converts class property to a python reference property constructor declaration.

    """
    return 'self.reference_to_{0} = {1}'.format(get_property_name(p), get_property_default_value(p))


def get_property_name(name):
    """Converts name to a python class property name.

    """
    return _strip(name)


def get_property_field_name(name):
    """Converts name to a python class property field name.

    """
    return _PROPERTY_FIELD_PREFIX + _strip(name)


def get_property_default_value(p):
    """Returns property default value.

    """
    # Return value based upon property type:
    # ... meta information;
    if p.name == "meta":
        if p.package == p.type.package:
            return "{0}()".format(get_type_name(p.type))
        else:
            return "{0}.{1}()".format(get_package_name(p.type.package),
                                      get_type_name(p.type))

    # ... collection types;
    elif p.is_collection:
        return _NULL_VALUE_COLLECTION

    # ... enum / complex / simple types.
    else:
        return _NULL_VALUE


def get_type_name(type):
    """Returns python type name.

    :param str name: Type name being converted.

    """
    name = type.name
    if type.is_simple:
        return _get_simple_type_mapping(name)
    elif type.is_enum:
        return _get_simple_type_mapping('unicode')
    elif type.is_complex:
        return get_type_name(name)


def get_type_functional_name(t, get_full_name=False):
    """Returns python type functional name.

    :param str name: Type name being converted.

    """
    name = t.name
    if t.is_simple:
        return _get_simple_type_mapping(name)
    elif t.is_enum:
        return 'unicode'
    elif t.is_complex:
        if get_full_name:
            return get_package_name(name) + "." + get_type_name(name)
        else:
            return get_type_name(name)


def get_type_doc_name(t):
    """Returns python type documentation name.

    :param str name: Type name being converted.

    """
    name = t.name
    if t.is_simple:
        return _get_simple_type_mapping(name)
    elif t.is_enum or t.is_complex:
        return '{0}.{1}'.format(get_package_name(name), get_type_name(name))


def get_type_name(name):
    """Converts name to a python type name.

    :param str name: Type name being converted.

    """
    return convert_to_camel_case(_strip_type_name(name))


def _get_type_file_name(name):
    """Converts name to a python class file name.

    :param str name: Class name being converted.

    """
    return "{}{}".format(_strip_type_name(name), FILE_EXTENSION)


def _get_simple_type_mapping(simple):
    """Returns matching simple type mapping.

    """
    return _SIMPLE_TYPE_MAPPINGS[simple]


def _strip_package_name(name):
    """Returns stripped package name.

    """
    name = _strip(name)
    if name.find('.') != -1:
        name = name.split('.')[0]
    return name


def get_package_init_file_name():
    """Returns python package init file name.

    :param str name: Package name being converted.

    """
    return _PACKAGE_INIT_FILE + FILE_EXTENSION



def get_package_module_name(name, prefix):
    """Returns a package module name by injecting package name into a template.

    :param str name: Package name.
    :param str prefix: A package name prefix, e.g. "validator".

    """
    return (prefix + "_for_{0}_package").format(get_package_name(name))


def get_package_module_file_name(name, prefix):
    """Returns a package module file name by injecting package name into a template.

    :param str name: Package name.
    :param str prefix: A package name prefix, e.g. "validator".

    """
    return get_package_module_name(name, prefix) + FILE_EXTENSION


def get_module_file_name(name):
    """Returns a module file name.

    :param str name: Module name.

    """
    return name + FILE_EXTENSION


def format(o):
    """Pythonizes ontology names.

    """
    o.op_name = get_ontology_name(o)
    o.op_version = get_ontology_version(o)

    for p in o.packages:
        p.op_name = get_package_name(p)

    for c in o.classes:
        c.op_base_name = _get_class_base_name(c)
        c.op_doc_string_name = get_type_doc_string_name(c)
        c.op_file_name = _get_type_file_name(c)
        c.op_full_name = get_full_type_name(c)
        c.op_func_name = get_type_func_name(c)
        c.op_import_name = get_type_import_name(c)
        c.op_name = get_type_name(c)

    for e in o.enums:
        e.op_doc_string_name = get_type_doc_string_name(e)
        e.op_file_name = _get_type_file_name(e)
        e.op_full_name = get_full_type_name(e)
        e.op_func_name = get_type_func_name(e)
        e.op_import_name = get_type_import_name(e)
        e.op_name = get_type_name(e)


def emit_package_imports(o):
    """Emit set of package imports.

    """
    def emit_import(p):
        return "import {} as {}{}".format(
            get_package_module_name(p, 'typeset'),
            p.op_name,
            gu.emit_line_return())

    return gu.emit(o.packages, emit_import)





# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_sofware_model_component.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.ModelComponent instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test document type.
DOC_TYPE = cim.v1.ModelComponent

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Model"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key

# Test representation file.
DOC_FILE = 'cim.1.software.ModelComponent-1.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = '7a2b64cc-03ca-11e1-a36a-00163e9152a5'

# Test document version.
DOC_VERSION = '1'

# Test document creation date.
DOC_DATE = arrow.get('2012-04-23 14:59:06.258353').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_meta_info(meta):
    """Asserts a document's meta-information.

    :param object meta: Document meta information.

    """
    tu.assert_iter(meta.genealogy.relationships, 1, cim.v1.DocRelationship)

    r = meta.genealogy.relationships[0]
    tu.assert_str(r.description, "The HadGEM2-A model", True)
    tu.assert_str(r.direction, "toTarget")
    tu.assert_str(r.type, "previousVersionOf")
    tu.assert_str(r.target.reference.name, "HadGEM1")


def assert_extension_info(ext):
    """Asserts a document's extension information.

    :param object ext: Document extension information.

    """
    tu.assert_str(ext.display_name, "HadGEM2-A")
    tu.assert_str(ext.description, "The HadGEM2-A model", True)
    tu.assert_str(ext.full_display_name, "CMIP5 Model : MOHC - HadGEM2-A")
    tu.assert_int(ext.summary_fields, 2)
    tu.assert_str(ext.summary_fields[0], "HadGEM2-A")
    tu.assert_str(ext.summary_fields[1], "Hadley Global Environment Model 2 - Atmosphere")


def _assert_doc_core(doc, is_update):
    """Assert core information."""
    tu.assert_str(doc.description, "The HadGEM2-A model", True)
    if not is_update:
        tu.assert_str(doc.long_name, "Hadley Global Environment Model 2 - Atmosphere")
    tu.assert_date(doc.release_date, "2009")
    tu.assert_str(doc.short_name, "HadGEM2-A")
    tu.assert_str(doc.type, "model")
    tu.assert_int(doc.types, 2)
    tu.assert_str(doc.types[1], "model")


def _assert_doc_citations(doc, is_update):
    """Assert citations."""
    tu.assert_int(doc.citations, 2)

    c = doc.citations[0]
    tu.assert_str(c.collective_title, "Bellouin N., O. Boucher", True)
    tu.assert_str(c.location, "http://www.metoffice.gov.uk/publications/HCTN/HCTN_73.pdf")
    tu.assert_str(c.title, "Bellouin et al. 2007")
    tu.assert_str(c.type, "Online Other")


def _assert_doc_responsible_parties(doc, is_update):
    """Assert responsible parties."""
    tu.assert_int(doc.responsible_parties, 4)

    rp = doc.responsible_parties[0]
    tu.assert_str(rp.abbreviation, "Chris Jones")
    tu.assert_str(rp.address, "Met Office Hadley Centre", True)
    tu.assert_str(rp.email, "chris.jones@metoffice.gov.uk")
    tu.assert_str(rp.individual_name, "Chris Jones")
    tu.assert_str(rp.role, "PI")
    tu.assert_str(rp.url, r"http://www.metoffice.gov.uk/research/our-scientists/climate-chemistry-ecosystems/chris-jones", True)

    rp = doc.responsible_parties[1]
    tu.assert_str(rp.organisation_name, "UK Met Office Hadley Centre")


def _assert_doc_sub_components(doc, is_update):
    """Assert sub-components."""
    tu.assert_int(doc.sub_components, 4)

    sc = doc.sub_components[0]
    tu.assert_int(sc.citations, 1)
    tu.assert_str(sc.description, "The model includes interactive schemes", True)
    tu.assert_uuid(sc.meta.id, "7a44cb24-03ca-11e1-a36a-00163e9152a5")
    tu.assert_int(sc.meta.version, 1)
    tu.assert_date(sc.meta.create_date, "2012-04-23 14:59:04.757315")
    tu.assert_str(sc.long_name, "Aerosols")
    tu.assert_int(sc.properties, 4)
    tu.assert_int(sc.responsible_parties, 4)
    tu.assert_int(sc.sub_components, 3)
    tu.assert_str(sc.short_name, "Aerosols")
    tu.assert_str(sc.type, "Aerosols")
    tu.assert_int(sc.types, 3)

    # Property.
    p = sc.properties[0]
    tu.assert_bool(p.is_represented, True)
    tu.assert_str(p.long_name, "Aerosols")
    tu.assert_str(p.short_name, "Aerosol Key Properties")
    tu.assert_int(p.sub_properties, 6)

    p = sc.properties[2]
    tu.assert_int(p.standard_names, 1)
    tu.assert_str(p.standard_names[0], "tendency_of_atmosphere", True)
    tu.assert_str(p.units, "kg/m2/s")

    # Sub-property.
    p = sc.properties[0].sub_properties[0]
    tu.assert_str(p.short_name, "AerosolSchemeScope")
    tu.assert_str(p.long_name, "AerosolSchemeScope")
    tu.assert_int(p.values, 1)
    tu.assert_str(p.values[0], "Whole atmosphere")



def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    for assertor in (
        _assert_doc_core,
        _assert_doc_citations,
        _assert_doc_responsible_parties,
        _assert_doc_sub_components
        ):
        assertor(doc, is_update)


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    doc.long_name = "X"


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    tu.assert_str(doc.long_name, "X")


def assert_doc_extensions(doc):
    """Asserts a document after being extended.

    :param object doc: Document being tested.

    """
    pass



# -*- coding: utf-8 -*-

"""
.. module:: designing_numerical_experiment.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.designing.numerical_experiment document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc import constants



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_viewer_urls,
        _set_summary_fields,
        _set_full_display_name
        )


def _set_viewer_urls(ctx):
    """Sets related experiment viewer urls.

    """
    for i in ctx.doc.related_experiments + ctx.doc.related_mips:
        try:
            i.meta
        except AttributeError:
            i.viewer_url = constants.VIEWER_URL_BY_ID.format(ctx.meta.project, i.id, i.version)
        else:
            i.viewer_url = constants.VIEWER_URL_BY_ID.format(ctx.meta.project, i.meta.id, i.meta.version)


def _set_summary_fields(ctx):
    """Sets related experiment viewer urls.

    """
    ctx.ext.summary_fields += (
        ctx.doc.canonical_name,
        ctx.doc.name,
        ctx.doc.long_name
    )


def _set_full_display_name(ctx):
    """Sets document full display name.

    """
    ctx.ext.full_display_name = "{} Tier {} Experiment: {}".format(
        ctx.meta.project.upper(),
        ctx.doc.tier,
        ctx.ext.display_name
        ).strip()



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.schema_validation.enum_validator
   :platform: Unix, Windows
   :synopsis: Validates ontology enum definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import re



# Regular expressions.
_RE_ENUM_NAME = '^[a-z_]+$'
_RE_ENUM_MEMBER_NAME = '^[a-zA-Z0-9-_ \.\:+]+$'


def _validate_enum_member(ctx, module, factory, enum, name, doc_string):
    """Validates an enumeration member.

    """
    if not re.match(_RE_ENUM_MEMBER_NAME, name):
        err = 'Invalid enum member: {0}.{1} --> name contain invalid characters'
        err = err.format(ctx.get_name(factory, module), name)
        ctx.set_error(err)

    if doc_string is not None and not len(doc_string.strip()):
        err = 'Invalid enum member: {0}.{1} --> doc string must be either None or a string'
        err = err.format(ctx.get_name(factory, module), name)
        ctx.set_error(err)


def validate(ctx, module, factory, enum):
    """Validates an enumeration.

    """
    if not re.match(_RE_ENUM_NAME, factory.__name__):
        err = 'Invalid enum: {} --> name format must be lower_case_underscore'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)

    if 'is_open' not in enum:
        err = 'Invalid enum: {} --> required attribute "is_open" is missing'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    if not isinstance(enum['is_open'], bool):
        err = 'Invalid enum: {} --> "is_open" attribute must be a boolean'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)

    if 'members' in enum and not isinstance(enum['members'], list):
        err = 'Invalid enum: {} --> "members" attribute must be a list'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    if [m for m in enum.get('members', []) if not isinstance(m, tuple) or not len(m) == 2]:
        err = 'Invalid enum: {} --> all members must be 2 item tuples (name, doc_string)'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    for m in enum.get('members', []):
        _validate_enum_member(ctx, module, factory, enum, m[0], m[1])



import pyesdoc



def pluck(section, item=None, attr=None):
	section = pyesdoc.drq.content[section]
	if item is None:
		return section

	item = section[item]
	if attr is None:
		return item

	return getattr(item, attr)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.utils.convert.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of library conversion functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections
import csv
import datetime
import json
import re
import time
import types
import uuid

import arrow



# Default character set.
_JSON_CHARSET = "ISO-8859-1"

# ISO date formats.
_ISO_DATE_FORMATS = ["%Y-%m-%d %H:%M:%S", "%Y-%m-%dT%H:%M:%S"]

# Values considered to be abbreviations.
_ABBREVIATIONS = ("id", "uid", "uuid")


def str_to_unicode(val):
    """Converts string input to a unicode literal.

    :param object val: arget to be converted to a unicode literal.

    :returns: A unicode literal.
    :rtype: unicode

    """
    if val is None:
        return unicode()
    if isinstance(val, unicode):
        return val

    val = str(val).decode('utf-8').strip()
    if not len(val):
        return unicode()

    return unicode(val)


def unicode_to_str(val):
    """Converts unicode input to a string literal.

    :param object val: arget to be converted to a unicode literal.

    :returns: A unicode literal.
    :rtype: unicode

    """
    if val is None:
        return str()

    if not isinstance(val, unicode):
        val = unicode(val)

    val = val.encode('utf-8').strip()
    if not len(val):
        return str()

    return val


def capitalize(target):
    """Capitalizes passed string.

    :param str target: A string to be converted.

    :returns: The target string capitalized.
    :rtype: str

    """
    if not isinstance(target, (str, unicode)):
        raise TypeError()

    if len(target) == 0:
        return target
    elif len(target) == 1:
        return target[0].upper()
    else:
        return target[0].upper() + target[1:]


def str_to_pascal_case(target, separator='_'):
    """Converts passed name to pascal case.

    :param target: A string to be converted.
    :type target: str

    :param separator: A separator used to split target string into constituent parts.
    :type separator: str

    :returns: The target string converted to pascal case.
    :rtype: str

    """
    r = ''
    if target is not None and len(target):
        # Preserve initial separator
        if target[0:len(separator)] == separator:
            r = separator

        # Iterate string parts.
        s = target.split(separator)
        for s in s:

            # Upper case abbreviations.
            if s.lower() in _ABBREVIATIONS:
                r += s.upper()

            # Upper case initial character.
            elif (len(s) > 0):
                r += s[0].upper()
                if (len(s) > 1):
                    r += s[1:]

    return r


def str_to_camel_case(target, separator='_'):
    """Converts passed name to camel case.

    :param target: A string to be converted.
    :type target: str

    :param separator: A separator used to split target string into constituent parts.
    :type separator: str

    :returns: The target string converted to camel case.
    :rtype: str

    """
    r = ''
    if target is not None and len(target):
        # Convert to pascal case.
        s = str_to_pascal_case(target, separator)

        # Preserve initial separator
        if s[0:len(separator)] == separator:
            r += separator
            s = s[len(separator):]

        # Lower case abbreviations.
        if s.lower() in _ABBREVIATIONS:
            r += s.lower()

        # Lower case initial character.
        elif len(s):
            r += s[0].lower()
            r += s[1:]

    return r


def str_to_spaced_case(target, separator='_'):
    """Helper function to convert a string value from camel case to spaced case.

    :param target: A string for conversion.
    :type target: str

    :returns: A string converted to spaced case.
    :rtype: str

    """
    if target is None:
        return ""
    elif separator is not None and len(target.split(separator)) > 1:
        return " ".join(target.split(separator))
    elif target.find(" ") == -1:
        return re.sub("([A-Z])"," \g<0>", target).strip()
    else:
        return target


def str_to_underscore_case(target):
    """Helper function to convert a from camel case string to an underscore case string.

    :param target: A string for conversion.
    :type target: str

    :returns: A string converted to underscore case, e.g. account_number.
    :rtype: str

    """
    if target is None or not len(target):
        return ''

    r = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', target)
    r = re.sub('([a-z0-9])([A-Z])', r'\1_\2', r)
    r = r.lower()

    return r


class _JSONEncoder(json.JSONEncoder):
    """Extends json encoder so as to handle extended types.

    """
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat().replace('T', ' ')
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.time):
            return obj.isoformat()
        elif isinstance(obj, uuid.UUID):
            return str(obj)
        else:
            raise TypeError(repr(obj) + " is not JSON serializable")


class _JSONDecoder(json.JSONDecoder):
    """Extends json decoder so as to handle extended types.

    """
    def __init__(self, key_formatter, to_namedtuple=False):
        json.JSONDecoder.__init__(self,
                                  encoding=_JSON_CHARSET,
                                  object_hook=self.dict_to_object)
        self.key_formatter = key_formatter
        self.to_namedtuple = to_namedtuple
        self.value_parsers = [
            self.unicode_to_datetime,
            self.unicode_to_uuid
            ]


    def dict_to_object(self, d):
        # Parse values.
        for k, v in d.items():
            for parser in self.value_parsers:
                if parser(d, k, v):
                    break

        # Format keys.
        if self.key_formatter is not None:
            d = dict_keys(d, self.key_formatter)

        # Return dictionary | named tuple.
        return d if not self.to_namedtuple else dict_to_namedtuple(d)


    def unicode_to_datetime(self, d, k, v):
        if isinstance(v, unicode) and len(v):
            try:
                float(v)
            except ValueError:
                for format in _ISO_DATE_FORMATS:
                    try:
                        v = datetime.datetime.strptime(v, format)
                    except (ValueError, TypeError):
                        pass
                    else:
                        d[k] = v
                        return True

        return False


    def unicode_to_uuid(self, d, k, v):
        if isinstance(v, unicode) and len(v):
            try:
                v = uuid.UUID(v)
            except ValueError:
                pass
            else:
                d[k] = v
                return True

        return False


def json_to_dict(repr, key_formatter=None):
    """Converts a json encoded string to a dictionary.

    :param repr: A json encoded string.
    :type repr: str | unicode

    :param key_formatter: Dictionary key formatter.
    :type key_formatter: function

    :returns: A dictionary.
    :rtype: dict

    """
    return _JSONDecoder(key_formatter).decode(repr)


def json_to_namedtuple(repr, key_formatter=None):
    """Converts a json encoded string to a namedtuple.

    :param repr: A json encoded string.
    :type repr: str | unicode

    :param key_formatter: Dictionary key formatter.
    :type key_formatter: function

    :returns: A namedtuple.
    :rtype: namedtuple

    """
    return _JSONDecoder(key_formatter, to_namedtuple=True).decode(repr)


def _convert_file(fp, convertor, key_formatter=None):
    """Converts a file."""
    with open(fp, 'r') as f:
        return convertor(f.read(), key_formatter)


def json_file_to_dict(fp, key_formatter=None):
    """Converts a json file to a dictionary.

    :param fp: A json file path.
    :type fp: str

    :param key_formatter: Dictionary key formatter.
    :type key_formatter: function

    :returns: A dictionary.
    :rtype: dict

    """
    return _convert_file(fp, json_to_dict, key_formatter)


def json_file_to_namedtuple(fp, key_formatter=None):
    """Converts a json file to a namedtuple.

    :param fp: A json file path.
    :type fp: str

    :param key_formatter: Dictionary key formatter.
    :type key_formatter: function

    :returns: A namedtuple.
    :rtype: namedtuple

    """
    return _convert_file(fp, json_to_namedtuple, key_formatter)


def dict_to_json(d, key_formatter=None):
    """Converts a dictionary to json.

    :param d: A dictionary.
    :type d: dict

    :param key_formatter: Dictionary key formatter.
    :type key_formatter: function

    :returns: A json encoded string.
    :rtype: str

    """
    if key_formatter is not None:
        d = dict_keys(d, key_formatter)

    return _JSONEncoder().encode(d)


def dict_to_namedtuple(d, key_formatter=None):
    """Converts a dictionary to a named tuple.

    :param d: Dictionary for conversion.
    :type d: dict

    :param key_formatter: Dictionary key formatter.
    :type key_formatter: function

    :returns: A named tuple.
    :rtype: namedtuple

    """
    if key_formatter is not None:
        d = key_formatter(d)

    _Class = collections.namedtuple('_Class', d.keys())

    return _Class(**d)


def dict_keys(d, key_formatter=str_to_pascal_case):
    """Converts keys of a dictionary using the passed key formatter.

    :param d: A dictionary.
    :type d: dict

    :param key_formatter: A dictionary key formatter function pointer.
    :type key_formatter: function

    :returns: A dictionary with it's keys formatted accordingly.
    :rtype: dict

    """
    if not isinstance(d, dict):
        return d

    r = {}

    for k, v in d.items():
        if isinstance(v, dict):
            r[key_formatter(k)] = dict_keys(v, key_formatter)
        elif isinstance(v, types.ListType):
            r[key_formatter(k)] = map(lambda i: dict_keys(i, key_formatter), v)
        else:
            r[key_formatter(k)] = v

    return r


def dict_keys_to_lower_case(d):
    """Converts keys of a dictionary to lower case.

    :param d: A dictionary.
    :type d: dict

    :returns: A dictionary with it's keys formatted accordingly.
    :rtype: dict

    """
    return dict_keys(d, lambda k: k.lower())


def dict_keys_to_upper_case(d):
    """Converts keys of a dictionary to upper case.

    :param d: A dictionary.
    :type d: dict

    :returns: A dictionary with it's keys formatted accordingly.
    :rtype: dict

    """
    return dict_keys(d, lambda k: k.upper())


def dict_keys_to_camel_case(d):
    """Converts keys of a dictionary to camel case.

    :param d: A dictionary.
    :type d: dict

    :returns: A dictionary with it's keys formatted accordingly.
    :rtype: dict

    """
    return dict_keys(d, str_to_camel_case)


def dict_keys_to_pascal_case(d):
    """Converts keys of a dictionary to pascal case.

    :param d: A dictionary.
    :type d: dict

    :returns: A dictionary with it's keys formatted accordingly.
    :rtype: dict

    """
    return dict_keys(d, str_to_pascal_case)


def dict_keys_to_underscore_case(d):
    """Converts keys of a dictionary to underscore case.

    :param d: A dictionary.
    :type d: dict

    :returns: A dictionary with it's keys formatted accordingly.
    :rtype: dict

    """
    return dict_keys(d, str_to_underscore_case)


def now_to_timestamp(offset=0):
    """Returns a timestamp from datatime.datetime.now().

    """
    now = time.time()
    localtime = time.localtime(now)
    milliseconds = '%03d' % int((now - int(now)) * 1000)
    ts = time.strftime('%Y%m%d%H%M%S', localtime) + milliseconds

    return int(ts) + offset


def csv_file_to_dict(fp):
    """Converts a csv file to a list of dictionaries.

    :param str fp: A csv file path.

    :returns: A list of dictionaries.
    :rtype: list

    """
    return map(lambda r:r, csv.DictReader(open(fp)))


def csv_file_to_namedtuple(fp):
    """Converts a csv file to a list of named tuples.

    :param str fp: A csv file path.

    :returns: A list of named tuples.
    :rtype: list

    """
    return map(dict_to_namedtuple, csv.DictReader(open(fp)))


def csv_file_to_json(fp):
    """Converts a csv file to json.

    :param str fp: A csv file path.

    :returns: A json string.
    :rtype: unicode

    """
    return dict_to_json(csv_file_to_dict(fp))


def csv_file_to_json_file(fp, dest=None):
    """Converts a csv file to a json file.

    :param str fp: A csv file path.

    """
    if dest is None:
        dest = fp.split(".")[0] + ".json"
    with open(dest, 'w') as f:
        f.write(dict_to_json(csv_file_to_dict(fp)))


def text_to_typed_value(text, target_type):
    """Converts text to a typed value.

    :param str text: Text to be converted to a typed value.
    :param class type: Target type.

    :returns: A typed value.
    :rtype: object

    """
    # Encode text.
    if text is not None:
        text = str_to_unicode(text) if isinstance(text, str) else unicode(text)

    # Decode according to type:
    # ... date's
    if target_type in {datetime.datetime, datetime.date, datetime.time}:
        return arrow.get(text).datetime
    # ... uuid's
    elif target_type is uuid.UUID:
        return uuid.UUID(text)
    # ... boolean's
    elif target_type is bool:
        return text.lower() in {"yes", "true", "t", "1", "y"}
    elif target_type is unicode:
        return text
    # ... others
    else:
        try:
            return target_type(text)
        # ... exceptions
        except Exception as e:
            print "Scalar decoding error", text, target_type



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc._publishing.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes document publishing functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime
import urllib
import uuid

import requests

from pyesdoc import constants
from pyesdoc import _api_proxy as proxy
from pyesdoc._extensions import extend
from pyesdoc._serialization import decode
from pyesdoc._serialization import encode
from pyesdoc._validation import is_valid



# Set of target endpoints.
_EP_CREATE = "/2/document/create"
_EP_RETRIEVE = "/2/document/retrieve"
_EP_UPDATE = "/2/document/update"
_EP_DELETE = "/2/document/delete"


def retrieve(uid, version=constants.DOC_VERSION_LATEST):
    """Retrieves a document from web service.

    :param str|uuid.UUID uid: Document unique identifier.
    :param str|int version: Document version.

    :returns: A document from remote repository.
    :rtype: object | None

    """
    # Validate inputs.
    if not isinstance(uid, uuid.UUID):
        try:
            uid = uuid.UUID(uid)
        except ValueError:
            _throw_invalid_doc_id()
    if version != constants.DOC_VERSION_LATEST and \
       not isinstance(version, int):
        _throw_invalid_doc_version()

    # Invoke web-service.
    endpoint = _get_doc_endpoint(_EP_RETRIEVE, uid, version, constants.ENCODING_JSON)
    response = proxy.invoke(requests.get, endpoint)
    proxy.parse_response(response, error_on_404=False)

    # Process web-service response.
    if response.text:
        return extend(decode(response.json(), constants.ENCODING_DICT))


def publish(doc):
    """Publishes a document to web service.

    :param doc: Document being published.
    :type doc: object

    :returns: Updated document.
    :rtype: object

    """
    # Validate inputs.
    if doc is None:
        raise TypeError("Cannot publish null documents.")
    if not is_valid(doc):
        raise ValueError("Cannot publish invalid documents.")

    # Increment version / update date.
    doc.meta.version = doc.meta.version + 1
    doc.meta.update_date = datetime.datetime.now()

    # Invoke web-service.
    endpoint = _EP_UPDATE if doc.meta.version > 1 else _EP_CREATE
    payload = encode(doc, constants.ENCODING_JSON)
    verb = requests.put if doc.meta.version > 1 else requests.post
    response = proxy.invoke(verb, endpoint, data=payload)

    # Process web-service response.
    proxy.parse_response(response)


def unpublish(uid, version=constants.DOC_VERSION_ALL):
    """Unpublishes a document from web service.

    :param str|uuid.UUID uid: Document unique identifier.
    :param str|int version: Document version.

    """
    # Validate inputs.
    if not isinstance(uid, uuid.UUID):
        try:
            uid = uuid.UUID(uid)
        except ValueError:
            _throw_invalid_doc_id()
    if not version == constants.DOC_VERSION_ALL and \
       not version == constants.DOC_VERSION_LATEST and \
       not isinstance(version, int):
        _throw_invalid_doc_version()

    # Invoke web-service.
    endpoint = _get_doc_endpoint(_EP_DELETE, uid, version)
    response = proxy.invoke(requests.delete, endpoint)

    # Process web-service response.
    proxy.parse_response(response)


def _throw_invalid_doc_id():
    """Throws an error.

    """
    raise TypeError("Invalid document uid (must be an instance of uuid.UUID).")


def _throw_invalid_doc_version():
    """Throws an error.

    """
    raise TypeError("Invalid document version (must be either 'all', 'latest' or an integer.")


def _get_doc_endpoint(endpoint, uid, version, encoding=None):
    """Returns a document instance endpoint url.

    """
    params = {
        "document_id": uid,
        "document_version": version
    }
    if encoding:
        params["encoding"] = encoding

    return "{}?{}".format(endpoint, urllib.urlencode(params))
